@@ -1,14 +1,13 @@
 ;;; wisitoken-grammar-mode.el --- Major mode for editing WisiToken grammar files  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2017 - 2019  Free Software Foundation, Inc.
+;; Copyright (C) 2017 - 2020  Free Software Foundation, Inc.
 
 ;; Author: Stephen Leake <stephen_leake@stephe-leake.org>
 ;; Maintainer: Stephen Leake <stephen_leake@stephe-leake.org>
 ;; Keywords: languages
-;; Version: 1.0.3
-;; package-requires: ((wisi "2.2.1") (emacs "25.0") (mmm-mode "0.5.7"))
-
-;; no upstream url; just ELPA
+;; Version: 1.2.0
+;; package-requires: ((wisi "3.1.1") (emacs "25.0") (mmm-mode "0.5.7"))
+;; url: http://www.nongnu.org/ada-mode/
 
 ;; This file is part of GNU Emacs.
 
@@ -41,16 +40,24 @@
   :group 'languages)
 
 (defcustom wisitoken-grammar-process-parse-exec "wisitoken_grammar_mode_parse.exe"
+  ;; wisitoken_grammar.gpr uses .exe even on non-windows.
   "Name of executable to use for external process wisitoken-grammar parser,"
   :type 'string
   :group 'wisitoken-grammar)
 
 (defvar wisitoken-grammar-mode-syntax-table
   (let ((table (make-syntax-table)))
-    ;; see wisitoken-grammar-syntax-propertize for comment start
-    (modify-syntax-entry ?\n ">   " table)
-    (modify-syntax-entry ?= ".   " table) ;; default symbol
-    (modify-syntax-entry ?* ".   " table) ;; default symbol
+    (modify-syntax-entry ?\; ". 12" table) ;; ";;" comment start; default punctuation
+    (modify-syntax-entry ?\n ">   " table) ;; comment end
+    (modify-syntax-entry ?=  ".   " table) ;; default symbol
+    (modify-syntax-entry ?*  ".   " table) ;; default symbol
+
+    ;; WORKAROUND (see mmm github issue 130) Because mmm does not
+    ;; limit syntax-ppss, we can't set ?\' to string here; that sees
+    ;; ?\' in 'code' and 'action' blocks. In addition, we prefer
+    ;; font-lock-constant-face for tokens. So we handle ?\' in the
+    ;; parser.
+
     table))
 
 (defvar wisitoken-grammar-mode-map
@@ -61,9 +68,18 @@
     (define-key map "\C-c\C-f" 'wisi-show-parse-error)
     (define-key map "\C-c\C-m" 'wisitoken-grammar-mmm-parse)
     (define-key map [S-return] 'wisitoken-grammar-new-line)
+    (define-key map "\C-c`"    'ada-show-secondary-error)
     map
   )  "Local keymap used for wisitoken-grammar mode.")
 
+(define-key emacs-lisp-mode-map "\C-c\C-m" 'wisitoken-grammar-mmm-parse)
+
+(defvar wisitoken-grammar-mode-menu (make-sparse-keymap "Wisi-Grammar"))
+(easy-menu-define wisitoken-grammar-mode-menu wisitoken-grammar-mode-map "Menu keymap for Wisitoken Grammar mode"
+  '("Wisi-Grammar"
+    ["Goto declaration" xref-find-definitions t]
+    ["mmm-ify action"   wisitoken-grammar-mmm-parse t]))
+
 (cl-defstruct (wisitoken-grammar-parser (:include wisi-process--parser))
   ;; no new slots
   )
@@ -168,7 +184,7 @@
   "If in action, call `mmm-parse-region' on it."
   (interactive)
   (save-excursion
-    (let* ((begin (search-backward-regexp "%[(}]" nil t))
+    (let* ((begin (search-backward-regexp "%[({]" nil t))
 	   (end   (when begin (search-forward-regexp "[)}]%" nil t))))
       (when (and begin end)
 	(mmm-parse-region begin end)))
@@ -229,19 +245,14 @@ Otherwise insert a plain new line."
     (goto-char start)
     (save-match-data
       (while (re-search-forward
-	      (concat
-	       "\\(;;\\)"     ;; comment start
-	       "\\|\\(%\\[\\)" ;; regexp begin
-	       )
+	       "\\(%\\[\\)" ;; regexp begin
 	      end t)
 	(cond
 	 ((match-beginning 1)
-	  (put-text-property (match-beginning 1) (match-end 1) 'syntax-table '(11 . nil)))
-
-	 ((match-beginning 2)
-	  (let ((begin (match-beginning 2))
+	  (let ((begin (match-beginning 1))
 		(end (search-forward "]%")))
-	    ;; allow single quotes in regexp to not mess up the rest of the buffer
+	    ;; allow single quotes in regexp to not mess up the rest
+	    ;; of the buffer
 	    (put-text-property begin end 'syntax-table '(11 . nil))
 	    ))
 	 ))
@@ -314,22 +325,29 @@ Otherwise insert a plain new line."
   (wisi-xref-identifier-at-point))
 
 (cl-defmethod xref-backend-identifier-completion-table ((_backend (eql wisitoken-grammar)))
-  (wisi-xref-identifier-completion-table))
+  (wisi-names t nil))
 
 (cl-defmethod xref-backend-definitions ((_backend (eql wisitoken-grammar)) identifier)
-  (unless (and (string-match wisi-xref-ident-regexp identifier)
-	       (match-string 2 identifier))
-    ;; Identifier is from identifier-at-point; get line from completion table
-    (setq identifier (try-completion identifier (wisi-xref-identifier-completion-table)))
-    (unless (test-completion identifier (wisi-xref-identifier-completion-table))
-      (setq identifier (completing-read "decl: " (wisi-xref-identifier-completion-table) nil t identifier)))
-    (string-match wisi-xref-ident-regexp identifier))
+  (when (get-text-property 0 'xref-identifier identifier)
+    ;; Identifier is from identifier-at-point; find declaration in completion table
+    (let* ((table (wisi-names t nil))
+	   (temp (try-completion identifier table)))
+      (cond
+       ((or (null temp)
+	    (not (test-completion temp table)))
+	(setq identifier (completing-read "decl: " table nil t identifier)))
 
-  (let* ((ident (match-string 1 identifier))
-	 (line-str (match-string 2 identifier))
-	 (line (when line-str (string-to-number line-str))))
-    (when line
-      (list (xref-make ident (xref-make-file-location (buffer-file-name) line  0))))
+       (t
+	(setq identifier temp)))
+      ))
+
+  ;; Identifier is now from completion table, or nil
+  (when identifier
+    (string-match wisi-names-regexp identifier)
+    (list (xref-make
+	 (match-string 1 identifier)
+	 (xref-make-file-location
+	  (buffer-file-name) (string-to-number (match-string 2 identifier)) 0)))
     ))
 
 ;;; debug
@@ -344,7 +362,7 @@ Otherwise insert a plain new line."
 (define-derived-mode wisitoken-grammar-mode prog-mode "Wisi"
   "A major mode for Wisi grammar files."
   (set (make-local-variable 'syntax-propertize-function) 'wisitoken-grammar-syntax-propertize)
-  (syntax-ppss-flush-cache (point-min));; reparse with new function
+
   (set (make-local-variable 'parse-sexp-ignore-comments) t)
   (set (make-local-variable 'parse-sexp-lookup-properties) t)
 
